@@ -66,9 +66,9 @@
 
 use core::arch::{asm, global_asm};
 
-use super::allocate_obj_on_stack;
+use super::{allocate_obj_on_stack, push};
 use crate::stack::{Stack, StackPointer};
-use crate::unwind::{InitialFunc, TrapHandler};
+use crate::unwind::{asm_may_unwind_root, InitialFunc, StackCallFunc, TrapHandler};
 use crate::util::EncodedValue;
 
 // The exact set of reserved registers can vary depending on the target and the
@@ -209,6 +209,8 @@ cfg_if::cfg_if! {
 }
 
 pub const STACK_ALIGNMENT: usize = 8;
+pub const PARENT_LINK_OFFSET: usize = 0;
+pub type StackWord = u32;
 
 global_asm!(
     ".balign 4",
@@ -242,7 +244,7 @@ global_asm!(
     "str sp, [r1, #-8]!",
     // Switch to our stack while popping the padding and initial PC. This also
     // sets up r2 for the third argument to the initial function.
-    "add r2, r2, #12",
+    "adds r2, r2, #12",
     "mov sp, r2",
     // Set up the frame pointer to point at the parent link. This is needed for
     // the unwinding code below.
@@ -306,25 +308,61 @@ global_asm!(
     asm_function_end!("stack_init_trampoline"),
 );
 
+global_asm!(
+    // See stack_init_trampoline for an explanation of the assembler directives
+    // used here.
+    ".balign 4",
+    thumb_symbol_def!(),
+    asm_function_begin!("stack_call_trampoline"),
+    ".cfi_sections .debug_frame",
+    ".fnstart",
+    ".cfi_startproc",
+    cfi_signal_frame!(),
+    // At this point our register state contains the following:
+    // - SP points to the top of the parent stack.
+    // - FP holds its value from the parent context.
+    // - R2 is the function that should be called.
+    // - R1 points to the top of our stack.
+    // - R0 contains the argument to be passed to the function.
+    //
+    // Create a stack frame and point the frame pointer at it. We use both
+    // DWARF and EHABI CFI directives just like stack_init_trampoline.
+    concat!("push {{", fp!(), ", lr}}"),
+    concat!("mov ", fp!(), ", sp"),
+    concat!(".save {{", fp!(), ", lr}}"),
+    concat!(".setfp ", fp!(), ", sp, #0"),
+    concat!(".cfi_def_cfa ", fp!(), ", 8"),
+    ".cfi_offset lr, -4",
+    concat!(".cfi_offset ", fp!(), ", -8"),
+    // Switch to the new stack.
+    "mov sp, r1",
+    // Call the function pointer. The argument is already in the correct
+    // register for the function.
+    "blx r2",
+    // Switch back to the original stack by restoring from the frame pointer,
+    // then return.
+    concat!("mov sp, ", fp!()),
+    concat!("pop {{", fp!(), ", pc}}"),
+    ".cfi_endproc",
+    ".fnend",
+    asm_function_end!("stack_call_trampoline"),
+);
+
+// These trampolines use a custom calling convention and should only be called
+// with inline assembly.
 extern "C" {
-    fn stack_init_trampoline();
+    fn stack_init_trampoline(arg: EncodedValue, stack_base: StackPointer, stack_ptr: StackPointer);
     fn stack_init_trampoline_return();
+    #[allow(dead_code)]
+    fn stack_call_trampoline(arg: *mut u8, sp: StackPointer, f: StackCallFunc);
 }
 
 #[inline]
 pub unsafe fn init_stack<T>(stack: &impl Stack, func: InitialFunc<T>, obj: T) -> StackPointer {
-    #[inline]
-    unsafe fn push(sp: &mut usize, val: Option<usize>) {
-        *sp -= 4;
-        if let Some(val) = val {
-            *(*sp as *mut usize) = val;
-        }
-    }
-
     let mut sp = stack.base().get();
 
     // Initial function.
-    push(&mut sp, Some(func as usize));
+    push(&mut sp, Some(func as StackWord));
 
     // Placeholder for parent link.
     push(&mut sp, None);
@@ -337,7 +375,7 @@ pub unsafe fn init_stack<T>(stack: &impl Stack, func: InitialFunc<T>, obj: T) ->
     debug_assert_eq!(sp % STACK_ALIGNMENT, 0);
 
     // Entry point called by switch_and_link().
-    push(&mut sp, Some(stack_init_trampoline as usize));
+    push(&mut sp, Some(stack_init_trampoline as StackWord));
 
     // Add an 8-byte offset because switch_and_link() looks for the target PC
     // 8 bytes above the stack pointer.
@@ -603,3 +641,22 @@ pub unsafe fn setup_trap_trampoline<T>(
         cpsr_endian: cfg!(target_endian = "big"),
     }
 }
+
+/// This function executes a function on the given stack. The argument is passed
+/// through to the called function.
+#[inline]
+pub unsafe fn on_stack(arg: *mut u8, stack: impl Stack, f: StackCallFunc) {
+    // This is a bit subtle: because we use .cfi_signal_frame in the trampoline,
+    // the unwinder will look for unwinding information at the instruction
+    // after the return address. Normal compiler code generation does not
+    // expect this and may generate incorrect entries in the exception handling
+    // table. We work around this by adding a NOP instruction after the call.
+    asm_may_unwind_root!(
+        concat!("blx ", asm_mangle!("stack_call_trampoline")),
+        "nop",
+        in("r0") arg,
+        in("r1") stack.base().get(),
+        in("r2") f,
+        clobber_abi("C"),
+    )
+}
