@@ -128,14 +128,17 @@
 use core::arch::{asm, global_asm};
 use core::sync::atomic::{AtomicUsize, Ordering};
 
-use super::allocate_obj_on_stack;
+use super::{allocate_obj_on_stack, push};
 use crate::stack::{Stack, StackPointer, StackTebFields};
 use crate::unwind::{
-    asm_may_unwind_root, asm_may_unwind_yield, cfi_reset_args_size_yield, InitialFunc, TrapHandler,
+    asm_may_unwind_root, asm_may_unwind_yield, cfi_reset_args_size_yield, InitialFunc,
+    StackCallFunc, TrapHandler,
 };
 use crate::util::EncodedValue;
 
-pub(crate) const STACK_ALIGNMENT: usize = 4;
+pub const STACK_ALIGNMENT: usize = 4;
+pub const PARENT_LINK_OFFSET: usize = 0;
+pub type StackWord = u32;
 
 // On MinGW, we emit DWARF CFI information so that it can be used by GDB to
 // reconstruct a backtrace. This isn't used for runtime unwinding (since we
@@ -154,6 +157,7 @@ cfg_if::cfg_if! {
 }
 
 global_asm!(
+    ".balign 16",
     asm_function_begin!("stack_init_trampoline"),
     dwarf!(".cfi_startproc"),
     dwarf!(".cfi_signal_frame"),
@@ -222,9 +226,69 @@ global_asm!(
     asm_function_end!("stack_init_trampoline"),
 );
 
+global_asm!(
+    // See stack_init_trampoline for an explanation of the assembler directives
+    // used here.
+    ".balign 16",
+    asm_function_begin!("stack_call_trampoline"),
+    dwarf!(".cfi_startproc"),
+    dwarf!(".cfi_signal_frame"),
+    // At this point our register state contains the following:
+    // - ESP points to the top of the parent stack.
+    // - EBP holds its value from the parent context.
+    // - EAX is the function that should be called.
+    // - EDX points to the top of our stack.
+    // - ECX contains the argument to be passed to the function.
+    //
+    // Create a stack frame and point the frame pointer at it.
+    "push ebp",
+    "mov ebp, esp",
+    dwarf!(".cfi_def_cfa ebp, 8"),
+    dwarf!(".cfi_offset ebp, -8"),
+    // Save the TEB fields to the stack.
+    "push dword ptr fs:[0xf78]", // GuaranteedStackBytes
+    "push dword ptr fs:[0xe0c]", // DeallocationStack
+    "push dword ptr fs:[0x8]",   // StackLimit
+    "push dword ptr fs:[0x4]",   // StackBase
+    "push dword ptr fs:[0x0]",   // ExceptionList
+    // Switch to the new stack.
+    "mov esp, edx",
+    // Pop the TEB fields for our new stack.
+    "pop dword ptr fs:[0x4]",   // StackBase
+    "pop dword ptr fs:[0x8]",   // StackLimit
+    "pop dword ptr fs:[0xe0c]", // DeallocationStack
+    "pop dword ptr fs:[0xf78]", // GuaranteedStackBytes
+    // At this point ESP points to the initial SEH exception handler record.
+    // Set it up in the TEB.
+    "mov fs:[0x0], esp",
+    // Call the function pointer. The argument is already in the correct
+    // register for the function.
+    "call eax",
+    // Save the two mutable TEB fields to the stack base so they can be
+    // retrieved later.
+    "push dword ptr fs:[0x8]",   // StackLimit
+    "push dword ptr fs:[0xf78]", // GuaranteedStackBytes
+    // Switch back to the original stack by restoring from the frame pointer,
+    // then return.
+    "lea esp, [ebp - 20]",
+    "pop dword ptr fs:[0x0]",   // ExceptionList
+    "pop dword ptr fs:[0x4]",   // StackBase
+    "pop dword ptr fs:[0x8]",   // StackLimit
+    "pop dword ptr fs:[0xe0c]", // DeallocationStack
+    "pop dword ptr fs:[0xf78]", // GuaranteedStackBytes
+    "pop ebp",
+    "ret",
+    dwarf!(".cfi_endproc"),
+    asm_function_end!("stack_call_trampoline"),
+);
+
+// These trampolines use a custom calling convention and should only be called
+// with inline assembly.
 extern "C" {
-    fn stack_init_trampoline();
+    fn stack_init_trampoline(arg: EncodedValue, stack_base: StackPointer, stack_ptr: StackPointer);
     fn stack_init_trampoline_return();
+    #[allow(dead_code)]
+    fn stack_call_trampoline(arg: *mut u8, sp: StackPointer, f: StackCallFunc);
 }
 
 /// The end of the exception handler chain is marked with 0xffffffff.
@@ -277,21 +341,13 @@ fn ntdll_final_exception_handler() -> usize {
 
 #[inline]
 pub unsafe fn init_stack<T>(stack: &impl Stack, func: InitialFunc<T>, obj: T) -> StackPointer {
-    #[inline]
-    unsafe fn push(sp: &mut usize, val: Option<usize>) {
-        *sp -= 4;
-        if let Some(val) = val {
-            *(*sp as *mut usize) = val;
-        }
-    }
-
     let mut sp = stack.base().get();
 
     // Placeholder for returning TEB.StackLimit.
     push(&mut sp, None);
 
     // Initial function.
-    push(&mut sp, Some(func as usize));
+    push(&mut sp, Some(func as StackWord));
 
     // Set up a proper "fake" return address for the frame pointer chain. We
     // just point it back at stack_init_trampoline_return to ensure that the
@@ -301,15 +357,15 @@ pub unsafe fn init_stack<T>(stack: &impl Stack, func: InitialFunc<T>, obj: T) ->
     // This is only needed on x86 Windows because this is the only platform
     // where the stack walker will alternate between looking at the frame
     // pointer chain and looking up FPO unwinding information in a PDB.
-    push(&mut sp, Some(stack_init_trampoline_return as usize));
+    push(&mut sp, Some(stack_init_trampoline_return as StackWord));
 
     // Placeholder for parent link.
     push(&mut sp, None);
 
     // Set up the initial SEH exception handler record. This must point to
     // FinalExceptionHandler in ntdll.dll for compatibility with SEHOP.
-    push(&mut sp, Some(ntdll_final_exception_handler()));
-    push(&mut sp, Some(EXCEPTION_LIST_END));
+    push(&mut sp, Some(ntdll_final_exception_handler() as StackWord));
+    push(&mut sp, Some(EXCEPTION_LIST_END as StackWord));
     let initial_seh_handler = sp;
 
     // Allocate space on the stack for the initial object, rounding to
@@ -318,18 +374,18 @@ pub unsafe fn init_stack<T>(stack: &impl Stack, func: InitialFunc<T>, obj: T) ->
     let initial_obj = sp;
 
     // Set up the stack for the 3rd argument to the initial function.
-    push(&mut sp, Some(initial_obj));
+    push(&mut sp, Some(initial_obj as StackWord));
 
     // Write the TEB fields for the target stack.
     let teb = stack.teb_fields();
-    push(&mut sp, Some(teb.GuaranteedStackBytes));
-    push(&mut sp, Some(teb.DeallocationStack));
-    push(&mut sp, Some(teb.StackLimit));
-    push(&mut sp, Some(teb.StackBase));
-    push(&mut sp, Some(initial_seh_handler));
+    push(&mut sp, Some(teb.GuaranteedStackBytes as StackWord));
+    push(&mut sp, Some(teb.DeallocationStack as StackWord));
+    push(&mut sp, Some(teb.StackLimit as StackWord));
+    push(&mut sp, Some(teb.StackBase as StackWord));
+    push(&mut sp, Some(initial_seh_handler as StackWord));
 
     // Entry point called by switch_and_link().
-    push(&mut sp, Some(stack_init_trampoline as usize));
+    push(&mut sp, Some(stack_init_trampoline as StackWord));
 
     StackPointer::new_unchecked(sp)
 }
@@ -632,8 +688,8 @@ pub unsafe fn drop_initial_obj(
 
     // Also copy the TEB fields to the base of the stack so that they can be
     // retreived by update_stack_teb_fields().
-    let base = stack_base.get() as *mut usize;
-    let stack = stack_ptr.get() as *const usize;
+    let base = stack_base.get() as *mut StackWord;
+    let stack = stack_ptr.get() as *const StackWord;
     *base.sub(1) = *stack.add(3); // StackLimit
     *base.sub(2) = *stack.add(5); // GuaranteedStackBytes
 }
@@ -644,9 +700,9 @@ pub unsafe fn drop_initial_obj(
 /// another coroutine.
 #[inline]
 pub unsafe fn update_stack_teb_fields(stack: &mut impl Stack) {
-    let base = stack.base().get() as *const usize;
-    let stack_limit = *base.sub(1);
-    let guaranteed_stack_bytes = *base.sub(2);
+    let base = stack.base().get() as *const StackWord;
+    let stack_limit = *base.sub(1) as usize;
+    let guaranteed_stack_bytes = *base.sub(2) as usize;
     stack.update_teb_fields(stack_limit, guaranteed_stack_bytes);
 }
 
@@ -654,12 +710,12 @@ pub unsafe fn update_stack_teb_fields(stack: &mut impl Stack) {
 /// saved on the parent stack.
 #[inline]
 pub unsafe fn read_parent_stack_teb_fields(stack_ptr: StackPointer) -> StackTebFields {
-    let stack_ptr = stack_ptr.get() as *const usize;
+    let stack_ptr = stack_ptr.get() as *const StackWord;
     StackTebFields {
-        StackBase: *stack_ptr.add(4),
-        StackLimit: *stack_ptr.add(5),
-        DeallocationStack: *stack_ptr.add(6),
-        GuaranteedStackBytes: *stack_ptr.add(7),
+        StackBase: *stack_ptr.add(4) as usize,
+        StackLimit: *stack_ptr.add(5) as usize,
+        DeallocationStack: *stack_ptr.add(6) as usize,
+        GuaranteedStackBytes: *stack_ptr.add(7) as usize,
     }
 }
 
@@ -671,9 +727,9 @@ pub unsafe fn update_parent_stack_teb_fields(
     stack_limit: usize,
     guaranteed_stack_bytes: usize,
 ) {
-    let stack_ptr = stack_ptr.get() as *mut usize;
-    *stack_ptr.add(5) = stack_limit;
-    *stack_ptr.add(7) = guaranteed_stack_bytes;
+    let stack_ptr = stack_ptr.get() as *mut StackWord;
+    *stack_ptr.add(5) = stack_limit as StackWord;
+    *stack_ptr.add(7) = guaranteed_stack_bytes as StackWord;
 }
 
 /// Registers which must be updated upon return from a trap handler.
@@ -718,8 +774,7 @@ pub unsafe fn setup_trap_trampoline<T>(
     let val_ptr = sp;
 
     // Set up a return address which returns to stack_init_trampoline.
-    sp -= 4;
-    *(sp as *mut u32) = stack_init_trampoline_return as u32;
+    push(&mut sp, Some(stack_init_trampoline_return as StackWord));
 
     // Set up registers for entry into the function.
     TrapHandlerRegs {
@@ -731,6 +786,45 @@ pub unsafe fn setup_trap_trampoline<T>(
     }
 }
 
+/// This function executes a function on the given stack. The argument is passed
+/// through to the called function.
+#[inline]
+pub unsafe fn on_stack<S: Stack>(arg: *mut u8, stack: S, f: StackCallFunc) {
+    let stack = scopeguard::guard(stack, |mut stack| {
+        let base = stack.base().get() as *const usize;
+        let stack_limit = *base.sub(4);
+        let guaranteed_stack_bytes = *base.sub(5);
+        stack.update_teb_fields(stack_limit, guaranteed_stack_bytes);
+    });
+
+    // Set up the values on the new stack.
+    let mut sp = stack.base().get();
+
+    // Add 4 bytes of padding for SAFESEH which requires that the exception
+    // handler record does not extend all the way to the stack base.
+    push(&mut sp, None);
+
+    // Set up the initial SEH exception handler record. This must point to
+    // FinalExceptionHandler in ntdll.dll for compatibility with SEHOP.
+    push(&mut sp, Some(ntdll_final_exception_handler() as StackWord));
+    push(&mut sp, Some(EXCEPTION_LIST_END as StackWord));
+
+    // Write the TEB fields for the target stack.
+    let teb = stack.teb_fields();
+    push(&mut sp, Some(teb.GuaranteedStackBytes as StackWord));
+    push(&mut sp, Some(teb.DeallocationStack as StackWord));
+    push(&mut sp, Some(teb.StackLimit as StackWord));
+    push(&mut sp, Some(teb.StackBase as StackWord));
+
+    asm_may_unwind_root!(
+        concat!("call ", asm_mangle!("stack_call_trampoline")),
+        in("ecx") arg,
+        in("edx") sp,
+        in("eax") f,
+        clobber_abi("fastcall"),
+    );
+}
+
 /// The trap handler will have reset our stack offset back to the base of the
 /// coroutine stack, but it won't have reset the SEH ExceptionList chain in the
 /// TIB. We need to manually reset it here before executing any user code which
